from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .max_pool import MaxPool
from .avg_pool import AvgPool

__all__ = [
    'MaxPool',
    'AvgPool'
]